from collections import OrderedDict
from contextlib import contextmanager

class ProgramExistsError(Exception): pass

class SupervisorConfator(object):
    
    def __init__(self, program_options={}, command_formats={}):
        self.programs = OrderedDict()
        self.groups = OrderedDict()
        self.program_options = program_options
        self.command_formats = command_formats
        self.context_options = dict()
        self.context_group_programs = []
    
    @contextmanager
    def options(self, **kwargs):
        self.context_options = kwargs
        yield
        self.context_options = {}
    
    @contextmanager
    def group(self, group_name, **kwargs):
        self.context_group_programs = []
        yield
        self.groups[group_name] = dict(programs=self.context_group_programs, **kwargs)
    
    def program(self, program_name, command, extra_command_formats={}, **kwargs):
        
        if program_name in self.programs:
            raise ProgramExistsError("Program '%s' already exists." % program_name)

        # Create the program options
        options = dict(**self.program_options)
        options.update(**self.context_options)
        options.update(**kwargs)
        
        # Format the command
        command_formats = dict(**self.command_formats)
        command_formats.update(**extra_command_formats)
        options.update(command=command.format(**command_formats))

        # Add the program        
        self.programs[program_name] = dict(**options)
        
        # Append it to the current group
        self.context_group_programs.append(program_name)
    
    def write(self):
        """
        Generate the output.
        """
        conf = []
        
        # First off write the programs
        initial_option_keys = ['command', 'user', 'process_name']
        
        for program,options in self.programs.items():
            conf.append('[program:%s]' % program)
            
            for key in initial_option_keys:
                try:
                    conf.append('%s=%s' % (key, self.to_supervisor(options.pop(key))))
                except KeyError: pass
            
            # Put the rest of the options in alpha order            
            for option,value in self.sort_options(options).items():
                conf.append('%s=%s' % (option, self.to_supervisor(value)))
        
            conf.append('')
        
        # Now write the groups
        for group,options in self.groups.items():
            conf.append('[group:%s]' % group)
            conf.append('programs=%s' % self.to_supervisor(options.pop('programs')))
            
            # Put the rest of the options in alpha order            
            for option,value in self.sort_options(options).items():
                conf.append('%s=%s' % (option, self.to_supervisor(value)))
        
            conf.append('')
            
        return "\n".join(conf)

    def sort_options(self, options):
        return OrderedDict(sorted(options.items(), key=lambda t: t[0]))

    def to_supervisor(self, value):
        if value is False:
            return 'false'
        elif value is True:
            return 'true'
        elif isinstance(value, (list, tuple)) and not isinstance(value, basestring):
            return ",".join([u'%s' % v for v in value])
        else:
            return '%s' % value


from unittest.case import TestCase
from collections import OrderedDict
from supervisor_confator import SupervisorConfator, ProgramExistsError

class SupervisorConfCreatorTestCase(TestCase):
    
    def test_one_program(self):
        sc = SupervisorConfator()
        sc.program('eggs', '/usr/bin/eggs')
        actual = sc.write()
        expected = """[program:eggs]
command=/usr/bin/eggs
"""
        self.assertEqual(expected, actual)
    
    def test_program_exists(self):
        sc = SupervisorConfator()
        sc.program('eggs', '/usr/bin/eggs')
        self.assertRaises(ProgramExistsError, sc.program, 'eggs', '/usr/bin/eggs')
    
    def test_complex(self):
        
        sc = SupervisorConfator(program_options=dict(user='myuser'),
                                command_formats=dict(bin='/usr/bin/',
                                                     log_dir='/var/log/'))
        
        sc.program('eggs', '{bin}eggs')
        
        with sc.options(autorestart=False):
            sc.program('sausage', '{bin}sausage --log-dir={log_dir}sausage.log')
        
        with sc.group('mygroup', priority=999):
            with sc.options(priority=998):
                sc.program('silly', '{bin}silly')
                sc.program('walks', '{bin}walks {extra}',
                           user='myotheruser', 
                           process_name='%(program_name)s',
                           numprocs=1,
                           numprocs_start=0,
                           startsecs=1,
                           startretries=3,
                           exitcodes=[0,2],
                           stopsignal='TERM',
                           stopwaitsecs=10,
                           redirect_stderr=False,
                           stdout_logfile='AUTO',
                           stdout_logfile_maxbytes='50MB',
                           stdout_logfile_backups=10,
                           stdout_capture_maxbytes=0,
                           stdout_events_enabled=False,
                           stderr_logfile='AUTO',
                           stderr_logfile_maxbytes='50MB',
                           stderr_logfile_backups=10,
                           stderr_capture_maxbytes=0,
                           stderr_events_enabled=False,
                           directory='/path/to/myapp.com',
                           serverurl='AUTO',
                           autostart=False, 
                           autorestart=False,
                           extra_command_formats=dict(extra=1))
            
        actual = sc.write()
        
        expected = """[program:eggs]
command=/usr/bin/eggs
user=myuser

[program:sausage]
command=/usr/bin/sausage --log-dir=/var/log/sausage.log
user=myuser
autorestart=false

[program:silly]
command=/usr/bin/silly
user=myuser
priority=998

[program:walks]
command=/usr/bin/walks 1
user=myotheruser
process_name=%(program_name)s
autorestart=false
autostart=false
directory=/path/to/myapp.com
exitcodes=0,2
numprocs=1
numprocs_start=0
priority=998
redirect_stderr=false
serverurl=AUTO
startretries=3
startsecs=1
stderr_capture_maxbytes=0
stderr_events_enabled=false
stderr_logfile=AUTO
stderr_logfile_backups=10
stderr_logfile_maxbytes=50MB
stdout_capture_maxbytes=0
stdout_events_enabled=false
stdout_logfile=AUTO
stdout_logfile_backups=10
stdout_logfile_maxbytes=50MB
stopsignal=TERM
stopwaitsecs=10

[group:mygroup]
programs=silly,walks
priority=999
"""
        self.assertEqual(expected, actual)


from codecs import open
import os

from setuptools import setup, find_packages


ROOT = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(ROOT, 'VERSION')) as f:
    VERSION = f.read().strip()

setup(
    name='supervisor-confator',
    packages=find_packages(exclude=['contrib', 'docs', 'tests*']),
    license='MIT',
    version = VERSION,
    description = 'Python interface to generate supervisor configuration files.',
    author = 'Alex Hayes',
    author_email = 'alex@alution.com',
    url = 'https://github.com/alexhayes/supervisor-confator',
    download_url = 'https://github.com/alexhayes/supervisor-confator/tarball/%s' % VERSION,
    keywords = ['supervisor', 'config', 'generator', 'server management'],
   
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',
         
        # Indicate who your project is intended for
        'Intended Audience :: System Administrators',
        'Topic :: System :: Installation/Setup',
         
        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',
         
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2.7',
    ],
    
    include_package_data=True,
    
)